/** @file

hippodraw::CreateNTuple class implementation

Copyright (C)  2006  The Board of Trustees of The Leland Stanford
Junior University.  All Rights Reserved.

$Id: CreateNTuple.cxx,v 1.28 2007/07/02 18:17:16 pfkeb Exp $

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "CreateNTuple.h"

#include "CanvasWindow.h"
#include "WindowController.h"
#include "Inspector.h"
#include "QtFileDialog.h"
#include "QtSortedCheckListItem.h"

#include "qcombobox.h"
#include "qlineedit.h"
#include "qmessagebox.h"

#include "datasrcs/DataSourceController.h"
#include "datasrcs/NTupleController.h"
#include "datasrcs/NTuple.h"
#include "datasrcs/TupleCut.h"

#include "controllers/CutController.h"
#include "plotters/Cut2DPlotter.h"

#include "fits/FitsController.h"

#include <cassert>

using std::string;
using std::vector;

using namespace hippodraw;

CreateNTuple::
CreateNTuple ( QWidget * parent, const char * name )
  : CreateNTupleBase ( parent, name )
{
  updateNTupleIn();
  initListViews ();
  currentDataSourceChanged (0);
}


void
CreateNTuple::
updateNTupleIn ()
{
  const vector < string > & nt_vector
    = DataSourceController::instance() -> getNTupleNames ();
  if ( nt_vector.empty () ) {
    m_ntuple_in -> clear ();
    return;
  }

  unsigned int count = m_ntuple_in -> count ();
  if ( count == nt_vector.size () ) return;

  m_ntuple_in -> clear();
  vector < string > ::const_iterator first = nt_vector.begin();
  while ( first != nt_vector.end() ) {
    const string & name = *first++;
    m_ntuple_in->insertItem ( name.c_str() );
  }

}

void
CreateNTuple::
updateColumnList ()
{
  
  m_ColumnListView->clear();

  DataSource * current_ds = DataSourceController::instance() -> getCurrent();
  columnNumber = current_ds -> columns();

  for ( unsigned int i = 0; i < columnNumber; i++ )
    {
      QString dummy = QString ("%1").arg(i);
      const std::string & s = current_ds -> getLabelAt ( i );
      const QString label = s.c_str();
      QtSortedCheckListItem * item = 
	new QtSortedCheckListItem (m_ColumnListView,dummy);
      item->setText( 1, QString ("%1").arg(label));
    }
}

void
CreateNTuple::
updateCutList()
{
  m_CutListView->clear();
  m_cut_list.clear ();
  
  const std::vector < PlotterBase * > & cutlist = getDataCutList();
  std::vector < PlotterBase * >::const_iterator it = cutlist.begin();

  unsigned int i = 0;

  for ( ; it != cutlist.end(); ++ it )
    {
      PlotterBase * plotter = *it;
      CutPlotter * cp = dynamic_cast < CutPlotter * > (plotter);
      cp -> fillCutList ( m_cut_list );

      Range cutrange = cp ->  getCutRange();
      QString dummy = QString ("%1").arg(i);
      const std::string & label = cp -> getLabel(Axes::X);
      double min = cutrange.low();
      double max = cutrange.high();
      const std::string & inversion = (cp -> getCutInversion())?"Inverted":"";

      QtSortedCheckListItem * item = 
	new QtSortedCheckListItem (m_CutListView,dummy);
      item->setText( 1, QString("%1").arg ( label.c_str() ) );
      item->setText( 2, QString("%1").arg(min));
      item->setText( 3, QString("%1").arg(max));
      item->setText( 5, QString("%1").arg( inversion.c_str() ) );
      
      Cut2DPlotter * cp2 = dynamic_cast < Cut2DPlotter * > (cp);
      if (cp2) 
	{
	  item->setText( 4, QString("2D-X"));
	  
	  // For Cut2DPlotter, update the information for the Y axis
	  ++i;
	  Range cutrange = cp2 -> getCutRangeY();
	  QString dummy = QString("%1").arg(i);
	  const std::string & label = cp2-> getLabel(Axes::Y);
	  double min = cutrange.low();
	  double max = cutrange.high();

	  QtSortedCheckListItem * item 
	    = new QtSortedCheckListItem (m_CutListView,dummy);
	  item->setText( 1, QString("%1").arg ( label.c_str() ) );
	  item->setText( 2, QString("%1").arg(min));
	  item->setText( 3, QString("%1").arg(max));
	  item->setText( 4, QString("2D-Y"));
	  item->setText( 5, QString("%1").arg(inversion.c_str()));
	  ++i;
	}
      else
	{
	  item->setText( 4, QString ("1D"));
	  ++i;
	}
    }  
  cutNumber = i;

  assert ( m_cut_list.size() == cutNumber );
}

void
CreateNTuple::
initListViews()
{  
  m_CutListView->addColumn (QString ("Index"),20 );
  m_CutListView->addColumn (QString("Cut Name"), 20 );
  m_CutListView->addColumn (QString("Min"), 20 );
  m_CutListView->addColumn (QString("Max"), 20 );
  m_CutListView->addColumn (QString("2D Cut"), 5 );
  m_CutListView->addColumn (QString("Inversion"),5);

  m_ColumnListView->addColumn (QString("Index"), 20 );
  m_ColumnListView->addColumn (QString("Column Name"), 40 );

}

void
CreateNTuple::
currentDataSourceChanged ( int item )
{
  DataSourceController * controller = DataSourceController::instance ();
  controller -> setCurrentIndex ( item );
 
  // Set the ComboBox.
  m_ntuple_in-> setCurrentItem ( item );
  updateColumnList();
  updateCutList();

}

void
CreateNTuple::
getPlotterList ( std::vector < PlotterBase * > & plotterlist )
{
  plotterlist.clear();
  CanvasWindow * canvaz = WindowController::instance () -> currentCanvas();

  if ( canvaz !=0 ) {
    canvaz -> fillPlotterList ( plotterlist );
  }
}

const std::vector < PlotterBase * > &
CreateNTuple::
getDataCutList ( )
{
  vector < PlotterBase * > plotterlist;
  getPlotterList ( plotterlist );

  CutController * cutcontroller = CutController::instance();
  const DataSource * current_ds 
    = DataSourceController::instance() -> getCurrent();

  return cutcontroller->getCutList ( plotterlist, current_ds );
}


void
CreateNTuple::
createButtonClicked()
{
  DataSourceController * ds_controller = DataSourceController::instance();
  NTupleController * ntc = NTupleController::instance();
  DataSource * ds = ds_controller -> getCurrent();

  std::string name( (m_ntuple_out->text()).latin1() );

  if ( name.size()==0 ) name = "<no name>";  // default title

  setColumnList();
  setCutList();

  NTuple * nt = ntc->createNTuple (m_column_list, m_cut_list, ds);

  // Do nothing if no column or row selected.
  if ((nt->rows()==0) || (nt->columns()==0)) return;

  // Register the new NTuple.
  ntc-> registerNTuple ( name, nt );

  // Update the dialog and inspector.
  updateNTupleIn();
  WindowController::instance () -> getInspector() -> update();
}

void
CreateNTuple::
createFileButtonClicked()
{

  string filename = QtFileDialog::getExportTupleFilename ( this );

  if (filename.empty()==true) return;


  // Write data to the file.
  string::size_type pos = filename.find_last_of ('.' );
  const string suffix = filename.substr (pos);
  
  DataSourceController * ds_controller = DataSourceController::instance();
  DataSource * ds = ds_controller -> getCurrent();

  std::string name( (m_ntuple_out->text()).latin1() );

  if ( name.size()==0 ) name = "<no name>";  // default title

  setColumnList();
  setCutList();

  if ( suffix == ".fits" || suffix == ".gz" ) 
    {
#ifdef HAVE_CFITSIO
      FitsController * fc = FitsController::instance();
      // int retVal =
//       fc -> writeNTupleToFile ( m_column_list, m_cut_list, ds, filename, name);
      fc -> writeNTupleToFile ( ds, filename, name, m_column_list, m_cut_list);
#else
      QString message ( "Sorry, can not create FITS file.\n"
			"The application was not built with\n"
			"optional FITS support." );

      QMessageBox::critical ( this, // parent
			      "Create NTuple", // caption
			      message,
			      QMessageBox::Ok,
			      Qt::NoButton );
#endif
    }
  else
    {
      NTupleController * ntc = NTupleController::instance();
      // Need to deal with retVal;
      // int retVal = 
      ntc->createNTupleToFile (m_column_list, m_cut_list, ds, filename, name);
    }

  accept();

}

void
CreateNTuple::
cutListCheckBox_toggled ( bool yes )
{
#if QT_VERSION < 0x040000
  QListViewItem * item = m_CutListView -> firstChild();
#else
  Q3ListViewItem * item = m_CutListView -> firstChild();
#endif

  for (unsigned int i = 0; i<cutNumber; i++ )
    {

      QtSortedCheckListItem * item2 = 
	dynamic_cast < QtSortedCheckListItem * > (item);
      item2 -> setOn ( yes );      
      item = item->nextSibling();
    }
}

void
CreateNTuple::
columnListCheckBox_toggled ( bool yes )
{
#if QT_VERSION < 0x040000
  QListViewItem * item = m_ColumnListView -> firstChild();
#else
  Q3ListViewItem * item = m_ColumnListView -> firstChild();
#endif
  for (unsigned int i = 0; i<columnNumber; i++ )
    {
      
      QtSortedCheckListItem * item2 = 
	dynamic_cast < QtSortedCheckListItem * > (item);
      item2 -> setOn ( yes );     
      item = item->nextSibling();
    }
}

void
CreateNTuple::
setColumnList()
{

  m_column_list.clear();

#if QT_VERSION < 0x040000
  QListViewItem * item = m_ColumnListView -> firstChild();
#else
  Q3ListViewItem * item = m_ColumnListView -> firstChild();
#endif

  for (unsigned int i = 0; i<columnNumber; i++ )
    {

      QtSortedCheckListItem * item2 = 
	dynamic_cast < QtSortedCheckListItem * > (item);
      if (item2->isOn()) {

	std::string name( (item->text(1)).latin1() );
	m_column_list.push_back(name);
      }
      
      item = item->nextSibling();
    }
}

void CreateNTuple::
setCutList()
{
  // Keep newNTuple unchanged if there is no cuts.
  if ( cutNumber == 0 ) return;

  // First remove the cuts not used.
#if QT_VERSION < 0x040000
  QListViewItem * item = m_CutListView -> firstChild();
#else
  Q3ListViewItem * item = m_CutListView -> firstChild();
#endif
  vector < const TupleCut * >::iterator first = m_cut_list.begin ();
  while ( first != m_cut_list.end() ) {
    QtSortedCheckListItem * item2 
      = dynamic_cast < QtSortedCheckListItem * > ( item );      
    if ( item2 -> isOn () == false ) {
      first = m_cut_list.erase ( first );
    }
    else {
      first++;
    }
    item = item -> nextSibling ();
  }
}
